# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`RandomProvider`."""
import secrets

from .provider import Provider


class RandomProvider(Provider):
    name: str = 'random'

    async def random(self, length: int) -> bytes:
        """Return a random byte-sequence of the given length."""
        return secrets.token_bytes(nbytes=length)


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.type.char import UTF8String
from pyasn1.type.constraint import ValueSizeConstraint
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1.type.namedval import NamedValues
from pyasn1.type.univ import BitString
from pyasn1.type.univ import Integer
from pyasn1.type.univ import Sequence
from pyasn1.type.univ import SequenceOf
from pyasn1_modules.rfc2459 import MAX


class PKIFreeText(SequenceOf):
    componentType = UTF8String
    sizeSpec = SequenceOf.sizeSpec + ValueSizeConstraint(1, MAX)


class PKIStatus(Integer):
    namedValues = NamedValues(
        ('granted', 0),
        # when the PKIStatus contains the value zero a TimeStampToken, as
        # requested, is present.
        ('grantedWithMods', 1),
        # when the PKIStatus contains the value one a TimeStampToken,
        # with modifications, is present.
        ('rejection', 2), ('waiting', 3), ('revocationWarning', 4),
        # this message contains a warning that a revocation is
        # imminent
        ('revocationNotification', 5),
    )


class PKIFailureInfo(BitString):
    namedValues = NamedValues(
        # unrecognized or unsupported Algorithm Identifier
        ('badAlg', 0),
        # transaction not permitted or supported
        ('badRequest', 2),
        # the data submitted has the wrong format
        ('badDataFormat', 5),
        # the TSA's time source is not available
        ('timeNotAvailable', 14),
        # the requested TSA policy is not supported by the TSA
        ('unacceptedPolicy', 15),
        # the requested extension is not supported by the TSA
        ('unacceptedExtension', 16),
        # the additional information requested could not be understood
        # or is not available
        ('addInfoNotAvailable', 17),
        # the request cannot be handled due to system failure  }
        ('systemFailure', 25)
    )


class PKIStatusInfo(Sequence):
    componentType = NamedTypes(
        NamedType(
            name='status',
            asn1Object=PKIStatus()
        ),
        OptionalNamedType(
            name='statusString',
            asn1Object=PKIFreeText()
        ),
        OptionalNamedType(
            name='failInfo',
            asn1Object=PKIFailureInfo()
        )
    )

    @property
    def status_code(self) -> int:
        return self[0]

    @property
    def status(self) -> int:
        return self[1]


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Declares :class:`Signature`."""
import inspect
from typing import cast
from typing import Any

import pydantic

from ckms.core.exceptions import MissingProtectedClaim
from ckms.core.exceptions import MissingProtectedHeader
from ckms.types import IKeychain
from ckms.types import Verifier
from .joseheader_ import JOSEHeader
from .octetb64 import OctetB64


class Signature(pydantic.BaseModel):
    protected: bytes
    header: JOSEHeader
    signature: OctetB64

    @property
    def alg(self) -> str | None:
        return self.header.alg

    @property
    def kid(self) -> str | None:
        return self.header.kid

    @pydantic.root_validator(allow_reuse=True, pre=True)
    def preprocess(cls, values: dict[str, Any]) -> dict[str, Any]:
        protected = values.get('protected')
        if not protected:
            raise MissingProtectedHeader
        return values

    async def verify(
        self,
        keychain: IKeychain,
        payload: bytes,
        require_kid: bool = True
    ) -> bool:
        """Verifies that this signature was created for the given payload using the
        given `keychain`.
        
        The `require_kid` indicates if the ``kid`` is required to be present on in
        the JOSE header. If `require_kid` is ``True`` and the claim is not present,
        validation fails immediately. Otherwise, all keys in the `keychain` are
        attempted to verify the signature with the specified algorithm.
        """
        if not self.kid and require_kid:
            raise MissingProtectedClaim('kid')
        if not self.alg:
            raise MissingProtectedClaim('alg')
        result = False

        # Construct the signing input
        signing_input = b'.'.join([self.protected, payload])

        # Use the key identifier to lookup the key. If not present, resort to
        # all registered keys with the given algorithm.
        if self.kid and keychain.has(self.kid):
            spec = cast(Verifier, keychain.get(kid=self.kid))
            if spec.allows_algorithm(self.alg):
                result = spec.verify(
                    signature=self.signature,
                    message=signing_input
                )
                if inspect.isawaitable(result):
                    result = await result
        elif not require_kid:
            keychain = keychain.filter(
                algorithm=self.alg,
                kid=self.kid,
                use='sig',
                op='verify'
            )
            for spec in keychain.verifiers():
                if not spec.allows_algorithm(self.alg):
                    continue
                result = spec.verify(
                    signature=self.signature,
                    message=signing_input
                )
                if inspect.isawaitable(result):
                    result = await result
                if result:
                    break
            else:
                result = False
        return result



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.type import tag
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1.type.univ import Integer
from pyasn1.type.univ import Sequence


class Accuracy(Sequence):
    componentType = NamedTypes(
        OptionalNamedType(
            name='seconds',
            asn1Object=Integer()
        ),
        OptionalNamedType(
            name='millis',
            asn1Object=Integer().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        ),
        OptionalNamedType(
            name='micros',
            asn1Object=Integer().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        )
    )


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any

from pyasn1.type import namedtype
from pyasn1.type import univ
from pyasn1_modules.rfc2459 import AlgorithmIdentifier


class MessageImprint(univ.Sequence):
    componentType: namedtype.NamedTypes = namedtype.NamedTypes(
        namedtype.NamedType('hashAlgorithm', AlgorithmIdentifier()),
        namedtype.NamedType('hashedMessage', univ.OctetString())
    )

    @property
    def algorithm(self) -> univ.ObjectIdentifier:
        return self[0][0] # type: ignore

    @property
    def hashed_message(self) -> Any:
        return self[1] # type: ignore


"""Declares various models used by :class:`LocalProvider`."""
import pathlib
from typing import Any
from typing import Literal
from typing import TypeAlias

import pydantic
from cryptography.hazmat.primitives.ciphers import algorithms
from cryptography.hazmat.primitives.serialization import load_pem_private_key
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives.asymmetric import ed448
from cryptography.hazmat.primitives.asymmetric import ed25519
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric import x448
from cryptography.hazmat.primitives.asymmetric import x25519
from typing_extensions import Annotated

from ckms import types
from ckms.core import const
from ckms.core import models
from ckms.core import Provider
from ckms.types import KeyUseType


class HMAC(bytes):

    @property
    def key(self) -> bytes:
        return bytes(self)

    def __repr__(self) -> str:
        return f'<HMAC: {len(self)}>'


class LocalKey(pydantic.BaseModel):
    path: str | pathlib.Path
    password: bytes | None
    encoding: Literal['pem'] = 'pem'

    async def setup(self, spec: 'LocalKeySpecification') -> Any:
        with open(self.path, 'rb') as f:
            return load_pem_private_key(f.read(), self.password)


class ParameterLessKey(pydantic.BaseModel):

    async def setup(self, spec: 'LocalKeySpecification') -> Any:
        assert isinstance(spec.provider, types.IProvider) # nosec
        return await spec.provider.generate(
            types.GenerateKeyOperation(
                kty=spec.kty,
                curve=spec.curve,
                use=spec.use
            )
        )


class TransientKey(pydantic.BaseModel):
    length: int

    async def setup(self, spec: 'LocalKeySpecification') -> Any:
        assert isinstance(spec.provider, types.IProvider) # nosec
        return await spec.provider.generate(
            types.GenerateKeyOperation(
                kty=spec.kty,
                length=self.length,
                curve=spec.curve,
                use=spec.use
            )
        )


class ContentEncryptionKey(pydantic.BaseModel):
    cek: bytes

    async def setup(self, spec: 'LocalKeySpecification') -> algorithms.AES:
        return algorithms.AES(self.cek)


class Key(pydantic.BaseModel):
    __root__: LocalKey | TransientKey

    @classmethod
    def parse_obj(cls: type[pydantic.BaseModel], obj: Any) -> pydantic.BaseModel:
        return super().parse_obj(obj).__root__ # type: ignore

    async def setup(self, spec: 'LocalKeySpecification') -> 'Key':
        return await self.__root__.setup(spec)


class LocalKeySpecification(models.KeySpecification):
    kty: str | None = None # type: ignore
    provider: Literal['local'] | Provider = 'local' # type: ignore
    key: Any

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        values.update({
            'use': inspector.get_algorithm_use(values['algorithm'])
        })

    async def load(self) -> models.KeySpecification:
        """Load or generate the key as specified by the parameters."""
        assert isinstance(self.provider, types.IProvider)
        if not self.loaded:
            self.key = await self.key.setup(self)
        self.kid = self.provider.get_key_identifier(self)
        self.loaded = True
        return self

    def get_key_material(self) -> bytes:
        assert isinstance(self.provider, types.IProvider)
        return self.provider.inspector.to_pem(self.key)

    def has_key_material(self) -> bool:
        return True

    def get_private_key(self) -> Any:
        return self.key


class RSAKeySpecification(LocalKeySpecification):
    kty: Literal['RSA'] = 'RSA'
    #algorithm: types.RSAAlgorithmType | None = None
    use: KeyUseType | None = None
    key: LocalKey | TransientKey | rsa.RSAPrivateKey

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('use') and not values.get('algorithm'):
            raise ValueError(
                "Unable to infer the key algorithm. Specify either the "
                "`algorithm` or `use` parameter."
            )
        if values.get('use') == 'sig':
            values['algorithm'] = const.DEFAULT_RSA_SIGNING_ALG
        elif values.get('use') == 'enc':
            values['algorithm'] = const.DEFAULT_RSA_ENCRYPTION_ALG
        else:
            values['use'] = inspector.get_algorithm_use(values['algorithm'])

    def get_public_key(self) -> rsa.RSAPublicKey:
        assert isinstance(self.key, rsa.RSAPrivateKey)
        return self.key.public_key()


class EllipticCurveKeySpecification(LocalKeySpecification):
    kty: Literal['EC'] = 'EC'
    #algorithm: types.EllipticCurveAlgorithmType | None = None
    curve: types.EllipticCurveType | None = None
    use: KeyUseType | None = None
    key: LocalKey | ParameterLessKey = ParameterLessKey()

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('algorithm') and not values.get('use'):
            raise ValueError(
                "Specify either the `algorithm` parameter or the "
                "`use` parameter."
            )
        if values.get('algorithm'):
            values['use'] = inspector.get_algorithm_use(values['algorithm'])
        elif values.get('use') == 'sig' and not values.get('algorithm'):
            values['algorithm'] = const.DEFAULT_EC_SIGNING_ALGORITHM
        elif values.get('use') == 'enc' and not values.get('algorithm'):
            values['algorithm'] = const.DEFAULT_EC_ENCRYPTION_ALGORITHM
        else:
            raise ValueError("Unable to infer key parameters.")
        if not values.get('curve'):
            values['curve'] = inspector.get_algorithm_curve(values['algorithm'])

    def get_public_key(self) -> ec.EllipticCurvePublicKey:
        assert isinstance(self.key, ec.EllipticCurvePrivateKey)
        return self.key.public_key()


OKPPublic: TypeAlias = (
    ed448.Ed448PublicKey |
    ed25519.Ed25519PublicKey |
    x448.X448PublicKey |
    x25519.X25519PublicKey
)


OKPPrivate: TypeAlias = (
    ed448.Ed448PrivateKey |
    ed25519.Ed25519PrivateKey |
    x448.X448PrivateKey |
    x25519.X25519PrivateKey
)


class EdwardsCurveKeySpecification(LocalKeySpecification):
    kty: Literal['OKP'] = 'OKP'
    #algorithm: types.EdwardsCurveAlgorithmType | None
    curve: types.EdwardsCurveType | None
    key: LocalKey | ParameterLessKey = ParameterLessKey()
    use: KeyUseType

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('use') and not values.get('curve'):
            raise ValueError(
                "Specify either the `use` parameter the `curve` parameter."
            )
        if values.get('curve') in {'Ed448', 'Ed25519'}:
            values['use'] = 'sig'
        elif values.get('curve') in {'X448', 'X25519'}:
            values['use'] = 'enc'
        elif values.get('use') == 'sig':
            values['curve'] = const.DEFAULT_ED_SIGNING_CURVE
        elif values.get('use') == 'enc':
            values['curve'] = const.DEFAULT_ED_ENCRYPTION_CURVE
        else:
            raise ValueError("Unable to infer key parameters.")
        if values['use'] == 'sig':
            values['algorithm'] = 'EdDSA'
        if values['use'] == 'enc' and not values.get('algorithm'):
            values['algorithm'] = const.DEFAULT_ED_ENCRYPTION_ALGORITHM

    def get_public_key(self) -> OKPPublic:
        assert isinstance(self.key, OKPPrivate)
        return self.key.public_key()


class SymmetricKeySpecification(LocalKeySpecification):
    kty: Literal['oct'] = 'oct'
    #algorithm: types.HMACAlgorithmType | types.AESAlgorithmType | None
    use: KeyUseType | None = None
    key: ContentEncryptionKey | LocalKey | TransientKey | ParameterLessKey | None

    @pydantic.root_validator(pre=True)
    def set_key_params(cls, values: dict[str, Any]) -> dict[str, Any]:
        algorithm = values.get('algorithm')
        if algorithm is not None and not bool(values.get('key')):
            # TODO: Very ugly
            if str.startswith(algorithm, 'A128'):
                length = 16
            elif str.startswith(algorithm, 'A192'):
                length = 24
            elif str.startswith(algorithm, 'A256'):
                length = 32
            else:
                raise NotImplementedError(algorithm)
            values['key'] = {'length': length}
        return values

    @classmethod
    def autodiscover(
        cls,
        provider: types.IProvider,
        inspector: types.IKeyInspector,
        values: dict[str, Any]
    ) -> None:
        if not values.get('algorithm') and not values.get('use'):
            raise ValueError(
                "Specify either the `algorithm` parameter or the "
                "`use` parameter."
            )
        if values.get('algorithm'):
            values['use'] = inspector.get_algorithm_use(values['algorithm'])
        elif values.get('use') == 'sig':
            values['algorithm'] = const.DEFAULT_SYMMETRIC_SIGNING_ALGORITHM
        elif values.get('use') == 'enc':
            values['algorithm'] = const.DEFAULT_SYMMETRIC_ENCRYPTION_ALGORITHM

    def get_private_bytes(self) -> bytes:
        private = self.get_private_key()
        return private.key

    def get_private_key(self) -> algorithms.AES | HMAC:
        assert isinstance(self.key, (algorithms.AES, HMAC)), type(self.key)
        return self.key


class KeySpecification(models.BaseKeySpecification):
    __root__: Annotated[
        EllipticCurveKeySpecification |
        RSAKeySpecification |
        EdwardsCurveKeySpecification |
        SymmetricKeySpecification,
        pydantic.Field(discriminator='kty')
    ]

    @classmethod
    def generate(cls, **kwargs: Any) -> models.KeySpecification:
        return cls.parse_obj(kwargs)

    @classmethod
    def parse_obj(cls: type[pydantic.BaseModel], obj: Any) -> models.KeySpecification:
        return super().parse_obj(obj).__root__ # type: ignore


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1_modules.rfc2459 import Extensions
from pyasn1.type import tag
from pyasn1.type.namedtype import DefaultedNamedType
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1.type.namedval import NamedValues
from pyasn1.type.univ import Boolean
from pyasn1.type.univ import Integer
from pyasn1.type.univ import ObjectIdentifier
from pyasn1.type.univ import Sequence
from pyasn1.type.useful import GeneralizedTime

from .accuracy import Accuracy
from .generalname import GeneralName
from .messageimprint import MessageImprint


class TimestampInfo(Sequence):
    componentType = NamedTypes(
        NamedType(
            name='version',
            asn1Object=Integer(namedValues=NamedValues(('v1', 1)))
        ),
        OptionalNamedType(
            name='policy',
            asn1Object=ObjectIdentifier()
        ),
        NamedType(
            name='messageImprint',
            asn1Object=MessageImprint()
        ),
        # MUST have the same value as the similar field in
        # TimeStampReq
        NamedType(
            name='serialNumber',
            asn1Object=Integer()
        ),
        # Time-Stamping users MUST be ready to accommodate integers
        # up to 160 bits.
        NamedType(
            name='genTime',
            asn1Object=GeneralizedTime()
        ),
        OptionalNamedType(
            name='accuracy',
            asn1Object=Accuracy()
        ),
        DefaultedNamedType(
            name='ordering',
            asn1Object=Boolean(False)
        ),
        OptionalNamedType(
            name='nonce',
            asn1Object=Integer()
        ),
        # MUST be present if the similar field was present
        # in TimeStampReq.  In that case it MUST have the same value.
        OptionalNamedType(
            name='tsa',
            asn1Object=GeneralName().subtype(
                explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        ),
        OptionalNamedType(
            name='extensions',
            asn1Object=Extensions().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        )
    )

    @property
    def version(self) -> int:
        return int(self[0])

    @property
    def policy(self) -> str:
        return str(self[1])

    @property
    def message_imprint(self) -> MessageImprint:
        return self[2]

    @property
    def tsa(self):
        return self[8]



# Copyright 2022 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import secrets
from typing import cast
from typing import Any

import pydantic

from ckms.utils import current_timestamp
from .audiencetype import AudienceType
from .claimset import ClaimSet
from .invalidtoken import MaximumAgeExceeded
from .invalidtoken import TokenNotEffective
from .invalidtoken import TokenExpired
from .invalidtoken import WrongAudience


class JSONWebToken(ClaimSet):
    __module__: str = 'ckms.types'
    _strict_required: set[str] = {'jti', 'exp', 'iat', 'nbf'}
    iss: str | None = None
    sub: str | None = None
    aud: AudienceType = AudienceType()
    exp: int | None = None
    nbf: int | None = None
    iat: int | None = None
    jti: str | None = None

    @classmethod
    def fromjson(cls, payload: bytes | str) -> 'JSONWebToken':
        return cast(cls, super().fromjson(payload))

    @classmethod
    def frompayload(cls, payload: bytes | str) -> 'JSONWebToken':
        return cast(cls, super().frompayload(payload))

    @classmethod
    def strict(
        cls,
        *,
        iss: str,
        ttl: int,
        jti: str | None = None,
        **claims: Any
    ) -> 'JSONWebToken':
        now = current_timestamp()
        exp = now + ttl
        nbf = claims.setdefault('nbf', now)
        if nbf > exp:
            raise ValueError("The 'nbf' claim must not be in the future.")
        return cls.parse_obj({
            **claims,
            'iss': iss,
            'iat': now,
            'exp': exp,
            'jti': jti or secrets.token_urlsafe(24)
        })

    def add_audience(self, audience: str) -> None:
        """Adds an audience to the JSON Web Token (JWT)."""
        self.aud.add(audience)

    def expires_in(
        self,
        ttl: int | None = None,
        now: int | None = None
    ) -> int | None:
        """Set the time-to-live using the ``exp`` claim. Return
        the remaining seconds.
        """
        now = now if now is not None else current_timestamp()
        if ttl is not None:
            self.exp = now + ttl
        return (self.exp - now) if self.exp else None

    def is_selfsigned(self) -> bool:
        """Return a boolean indicating if the token is self-issuer."""
        return self.iss == self.sub and None not in {self.iss, self.sub}

    def verify(
        self,
        *,
        audience: set[str] | None = None,
        required: set[str] | None = None,
        now: int | None = None,
        max_age: int | None = None,
        strict: bool = False
    ) -> None:
        """Verifies the claims according to the specification laid out
        in the JOSE RFCs.
        """
        required = required or set()
        if max_age:
            required.add('iat')
        super().verify(required=required, strict=strict)
        now = now or current_timestamp()
        if self.exp is not None and self.exp <= now:
            raise TokenExpired
        if self.nbf and self.nbf > now:
            raise TokenNotEffective
        if audience and not bool(audience & self.aud):
            raise WrongAudience(allowed=audience)
        if max_age is not None and (now - cast(int, self.iat)) >= max_age:
            assert self.iat is not None
            raise MaximumAgeExceeded(max_age, now, self.iat)


class SecurityEventToken(JSONWebToken):
    """Represents a Security Event Token (SET as defined in :rfc:`8417`."""
    # RFC 8417 Security Event Token (SET)
    iss: str
    iat: int
    jti: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(24),
    )
    aud: AudienceType
    exp: None
    nbf: None
    events: dict[str, dict[str, Any]]
    toe: int | None
    txn: str | None


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.type import tag
from pyasn1.type.char import IA5String
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.univ import Any
from pyasn1.type.univ import Choice
from pyasn1.type.univ import ObjectIdentifier


class GeneralName(Choice):
    componentType = NamedTypes(
        NamedType(
            name='rfc822Name',
            asn1Object=IA5String().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1)
            )
        ),
        #namedtype.NamedType('dNSName', univ.Any().subtype(
        #    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 2))),
        #namedtype.NamedType('x400Address', univ.Any().subtype(
        #    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 3))),
        NamedType(
            name='directoryName',
            asn1Object=Any().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 4)
            )
        ),
        #namedtype.NamedType('ediPartyName', univ.Any().subtype(
        #    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 5))),
        #namedtype.NamedType('uniformResourceIdentifier', char.IA5String().subtype(
        #    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 6))),
        #namedtype.NamedType('iPAddress', univ.OctetString().subtype(
        #    implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 7))),
        NamedType(
            name='registeredID',
            asn1Object=ObjectIdentifier().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 8)
            )
        )
    )


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from cryptography import x509
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.serialization import Encoding
from pyasn1.codec.der import decoder
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1.type.univ import ObjectIdentifier
from pyasn1.type.univ import OctetString
from pyasn1.type.univ import Sequence
from pyasn1.type.univ import SetOf

from ckms.algorithm import Algorithm
from ckms.types import Digest
from ckms.types import JSONWebKey
from .status import PKIStatusInfo
from .timestamptoken import TimeStampToken


MessageDigestOID: ObjectIdentifier = ObjectIdentifier((1,2,840,113549,1,9,4,))
TimestampInfoOID: ObjectIdentifier = ObjectIdentifier((1,2,840,113549,1,9,16,1,4))


class TimestampResponse(Sequence):
    algorithms: dict[tuple[str, str], str] = {
        ('2.16.840.1.101.3.4.2.1', '1.2.840.113549.1.1.1'): 'RS256',
        ('2.16.840.1.101.3.4.2.2', '1.2.840.113549.1.1.1'): 'RS384',
        ('2.16.840.1.101.3.4.2.3', '1.2.840.113549.1.1.1'): 'RS512',
    }
    certificate: x509.Certificate
    componentType = NamedTypes(
        NamedType(
            name='status',
            asn1Object=PKIStatusInfo()
        ),
        OptionalNamedType(
            name='timeStampToken',
            asn1Object=TimeStampToken()
        )
    )

    @classmethod
    def frombytes(
        cls,
        certificate: x509.Certificate,
        response: bytes
    ) -> 'TimestampResponse':
        response, substrate = decoder.decode(response, asn1Spec=cls())
        if substrate:
            raise ValueError('Extra data returned')
        response.certificate = certificate
        return response

    @property
    def algorithm(self) -> Algorithm | None:
        name = self.algorithms.get(
            (
                self.digest_algorithm,
                str(self.signer['digestEncryptionAlgorithm'][0])
            )
        )
        return Algorithm.get(name)

    @property
    def content_type(self) -> ObjectIdentifier:
        return self.timestamp.content['contentInfo']['contentType']

    @property
    def digest_algorithm(self) -> str:
        return str(self.signer['digestAlgorithm'][0])

    @property
    def hasher(self) -> hashes.HashAlgorithm:
        return getattr(hashes, str.upper(Digest.digest_oid[self.digest_algorithm]))()

    @property
    def info(self):
        return self[0]

    @property
    def public_key(self) -> JSONWebKey:
        """A :class:`JSONWebKey` representing the public key of the TSA."""
        return JSONWebKey.fromcertificate(self.algorithm, self.certificate)

    @property
    def serial_number(self) -> int:
        """The serial number of the certificate that was used to sign the
        response.
        """
        return int(self.signer['issuerAndSerialNumber'][1])

    @property
    def signature(self) -> bytes:
        return bytes(self.signer['encryptedDigest'])

    @property
    def signed_data(self) -> bytes:
        # Validate only only signature for now
        content, _ = decoder.decode(
            substrate=self.timestamp.content['contentInfo']['content'],
            asn1Spec=OctetString()
        )
        digest = Digest.fromoid(
            data=bytes(content),
            oid=str(self.digest_algorithm)
        )
        signer = self.signer
        if bool(signer['authenticatedAttributes']):
            attrs = signer['authenticatedAttributes']
            for attr in attrs:
                if attr[0] != MessageDigestOID:
                    continue
                value, _ = decoder.decode(
                    substrate=attr[1][0],
                    asn1Spec=OctetString()
                )
                if bytes(value) != bytes(digest):
                    raise ValueError("Mismatch between content and attribute.")
                content = value
                break
            else:
                raise ValueError("No signed data present.")
        return bytes(content)

    @property
    def signers(self):
        return self.timestamp.content['signerInfos']

    @property
    def signer(self):
        return self.signers[0]

    @property
    def timestamp(self) -> TimeStampToken:
        return self[1]

    def validate(self) -> None:
        """Validates that the response is as expected."""
        if self.content_type != TimestampInfoOID:
            raise ValueError(f"Invalid content type: {str(self.content_type)}")
        if not self.signers:
            raise ValueError("The response did not contain a signature.")
        if len(self.signers) > 1:
            # The time-stamp token MUST NOT contain any signatures
            # other than the signature of the TSA.  The certificate
            # identifier (ESSCertID) of the TSA certificate MUST
            # be included as a signerInfo attribute inside a
            # SigningCertificate attribute (RFC 3161).
            raise ValueError(
                "The time-stamp token MUST NOT contain any signature "
                "other than the signature of the TSA."
            )
        if self.algorithm is None:
            raise ValueError(
                "The signing algorithm used by the TSA is not supported."
            )


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
import inspect
from typing import cast
from typing import Any

from cryptography import x509

from ckms.types import Digest
from ckms.types import Message
from ckms.types import IHTTPClient
from .types import TimestampRequest
from .types import TimestampResponse


class TimestampingAuthority:
    """Provides an interface to create a timestamp using a specific authority."""
    __module__: str = 'ckms.ext.tsa'
    certificate: x509.Certificate
    digest: str
    url: str

    def __init__(
        self,
        *,
        url: str,
        crtpath: str,
        digest: str = 'sha1'
    ):
        self.certificate = x509.load_pem_x509_certificate(open(crtpath, 'rb').read())
        self.digest = digest
        self.url = url

    async def timestamp(
        self,
        *,
        client: Any,
        data: bytes | Message,
        digest: str | None = None
    ) -> bytes:
        """Create a timestamp of the given input data."""
        client = cast(IHTTPClient, client)
        if not isinstance(data, (Digest, Message)):
            data = Message(buf=data, digest=digest or self.digest)

        assert isinstance(data, Message)
        response = client.post(
            url=self.url,
            content=TimestampRequest.new(
                digest=await data.digest(),
                digestmod=digest or self.digest
            ),
            headers={
                'Content-Type': 'application/timestamp-query'
            }
        )
        if inspect.isawaitable(response):
            response = await response
        return TimestampResponse.frombytes(self.certificate, response.content)



# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.codec.der import encoder
from pyasn1.type.namedtype import DefaultedNamedType
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1.type.namedval import NamedValues
from pyasn1.type import tag
from pyasn1.type.univ import Boolean
from pyasn1.type.univ import Integer
from pyasn1.type.univ import ObjectIdentifier
from pyasn1.type.univ import Sequence
from pyasn1_modules.rfc2459 import Extensions

from .algorithmidentifier import AlgorithmIdentifier
from .messageimprint import MessageImprint


class TimestampRequest(Sequence):
    componentType: NamedTypes = NamedTypes(
        NamedType('version', Integer(namedValues=NamedValues(('v1', 1)))),
        NamedType('messageImprint', MessageImprint()),
        OptionalNamedType('reqPolicy', ObjectIdentifier()),
        OptionalNamedType('nonce', Integer()),
        DefaultedNamedType('certReq', Boolean(False)),
        OptionalNamedType(
            'extensions',
            Extensions().subtype(
                implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0)
            )
        )
    )

    @classmethod
    def new(cls, digest: bytes, digestmod: str) -> bytes:
        message = MessageImprint()
        message.setComponentByPosition(0, AlgorithmIdentifier.new(digestmod))
        message.setComponentByPosition(1, digest)
        request = cls()
        request.setComponentByPosition(0, 'v1')
        request.setComponentByPosition(1, message)
        request.setComponentByPosition(4)
        return encoder.encode(request)


# Copyright 2018 Cochise Ruhulessin
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# type: ignore
from pyasn1.type import tag
from pyasn1.type.namedtype import NamedType
from pyasn1.type.namedtype import NamedTypes
from pyasn1.type.namedtype import OptionalNamedType
from pyasn1_modules.rfc2315 import signedData
from pyasn1_modules.rfc2315 import ContentInfo
from pyasn1_modules.rfc2315 import SignedData
from pyasn1.codec.ber import decoder

from .timestampinfo import TimestampInfo


class TimeStampToken(ContentInfo):
    componentType = NamedTypes(
        NamedType('contentType', signedData),
        OptionalNamedType(
            name='content',
            asn1Object=SignedData().subtype(
                explicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatConstructed, 0)
            )
        )
    )

    @property
    def content(self):
        return self[1]

    @property
    def info(self) -> TimestampInfo:
        x, s = decoder.decode(bytes(self.content['contentInfo']['content']))
        if s:
            raise ValueError('Incomplete decoding')
        x, s = decoder.decode(bytes(x), asn1Spec=TimestampInfo())
        if s:
            raise ValueError('Incomplete decoding')
        return x